/*
 * BTN11 -- A simple PDP-11 assembler - it's better than nothing at all
 * By Nils M Holm, 2001. In the Public Domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAXERRORS	20
#define MAXCOND		8

#define SYMLEN		127
#define LISTBUFLEN	255
#define EMITBUFLEN	1024

#define ENDFILE		-1
#define NEWLINE		0077
#define AT		0100
#define LPAREN		0101
#define RPAREN		0102
#define COMMA		0103
#define PLUS		0104
#define MINUS		0105
#define COLON		0106
#define HASH		0107
#define EQUAL		0110
#define NUMBER		0200
#define SYMBOL		0201
#define STRING		0202
#define KEYWORD		0203
#define REGISTER	0204

struct mnemo {
	char	*sym;
	int	opcode;
	int	(*group)();
	int	flags;
};

struct syment {
	char		*name;
	int		addr;
	int		flags;
	struct syment	*next;
};

#define F_PREDEF	0001

char		*Infile = NULL, *Outfile = NULL;
FILE		*Input, *Output;
char		*Tmpfile = "btn11.tmp";
int		Token;
struct syment	*Symtab = NULL;
int		Value;
int		Radix = 8;
int		Rejected = 0;
char		Text[SYMLEN+1];
int		Line = 1;
int		Errors = 0;
int		Nkw;
int		Kw;
int		Reg;
int		Mode1, Reg1, Addr1,
		Mode2, Reg2, Addr2;
int		Cp, Lastcp;
int		Phase = 1;
char		Emitbuf[EMITBUFLEN];
int		Doemit = 1, Condstack[MAXCOND], Csp = 0;
int		Dolist = 0;
int		Donumber = 0;
int		Bootmode = 0;
char		Listbuf[LISTBUFLEN+1];
int		Lp = 0, Off = 0;
FILE		*Listfile;
int		Nosym;
int		Origin = 0;

int		Absmode = 0,
		Keeptmp = 0;

#define MREGISTER	0			/* Rx		*/
#define MDEFERRED	1			/* @Rx   (Rx)	*/
#define MINCREMENT	2			/* (Rx)+	*/
#define MINCREMENTDEF	3			/* @(Rx)+	*/
#define MDECREMENT	4			/* -(Rx)	*/
#define MDECREMENTDEF	5			/* @-(Rx)	*/
#define MINDEX		6			/* N(Rx)	*/
#define MINDEXDEF	7			/* @N(Rx)	*/
#define MIMMEDIATE	MINCREMENT		/* #N    (R7)+	*/
#define MABSOLUTE	MINCREMENTDEF		/* @#N   @(R7)+	*/
#define MRELATIVE	MINDEX			/* N(R7) N	*/
#define MRELATIVEDEF	MINDEXDEF		/* @N(R7)	*/

#define PC	7

#define P_ASCII		001
#define P_ASCIZ		002
#define P_BLKB		003
#define P_BLKW		004
#define P_BYTE		005
#define P_ELSE		006
#define P_ENDIF		007
#define P_EQU		010
#define P_EVEN		011
#define P_IF		012
#define P_ODD		013
#define P_ORG		014
#define P_RADIX		015
#define P_WORD		016

int	grp0(),		/* literal */
	grp1(),		/* single-operand */
	grp2(),		/* two-operand */
	grp3(),		/* operand-to-register */
	grp3b(),	/* register-to-operand */
	grp4(),		/* PC-relative */
	grp5(),		/* traps */
	mark(),		/* MARK */
	rts(),		/* RTS */
	sob(),		/* SOB */
	spl(),		/* SPL */
	pseudo();	/* pseudo instructions */

struct mnemo	Kwlist[] = {
	{ ".ASCII",	P_ASCII,	pseudo,	0	},
	{ ".ASCIZ",	P_ASCIZ,	pseudo,	0	},
	{ ".BLKB",	P_BLKB,		pseudo,	0	},
	{ ".BLKW",	P_BLKW,		pseudo,	0	},
	{ ".BYTE",	P_BYTE,		pseudo,	0	},
	{ ".ELSE",	P_ELSE,		pseudo,	0	},
	{ ".ENDIF",	P_ENDIF,	pseudo,	0	},
	{ ".EQU",	P_EQU,		pseudo,	0	},
	{ ".EVEN",	P_EVEN,		pseudo,	0	},
	{ ".IF",	P_IF,		pseudo,	0	},
	{ ".ODD",	P_ODD,		pseudo,	0	},
	{ ".ORG",	P_ORG,		pseudo,	0	},
	{ ".RADIX",	P_RADIX,	pseudo,	0	},
	{ ".WORD",	P_WORD,		pseudo,	0	},
	{ "ADC",	0005500,	grp1,	0	},
	{ "ADCB",	0105500,	grp1,	0	},
	{ "ADD",	0060000,	grp2,	0	},
	{ "ASH",	0072000,	grp3,	0	},
	{ "ASHC",	0073000,	grp3,	0	},
	{ "ASL",	0006300,	grp1,	0	},
	{ "ASLB",	0106300,	grp1,	0	},
	{ "ASR",	0006200,	grp1,	0	},
	{ "ASRB",	0106200,	grp1,	0	},
	{ "BCC",	0103000,	grp4,	0	},
	{ "BCS",	0103400,	grp4,	0	},
	{ "BEQ",	0001400,	grp4,	0	},
	{ "BGE",	0002000,	grp4,	0	},
	{ "BGT",	0003000,	grp4,	0	},
	{ "BHI",	0101000,	grp4,	0	},
	{ "BHIS",	0103000,	grp4,	0	},
	{ "BIC",	0040000,	grp2,	0	},
	{ "BICB",	0140000,	grp2,	0	},
	{ "BIS",	0050000,	grp2,	0	},
	{ "BISB",	0150000,	grp2,	0	},
	{ "BIT",	0030000,	grp2,	0	},
	{ "BITB",	0130000,	grp2,	0	},
	{ "BLE",	0003400,	grp4,	0	},
	{ "BLO",	0103400,	grp4,	0	},
	{ "BLOS",	0101400,	grp4,	0	},
	{ "BLT",	0002400,	grp4,	0	},
	{ "BMI",	0100400,	grp4,	0	},
	{ "BNE",	0001000,	grp4,	0	},
	{ "BPL",	0100000,	grp4,	0	},
	{ "BPT",	0000003,	grp0,	0	},
	{ "BR",		0000400,	grp4,	0	},
	{ "BVC",	0102000,	grp4,	0	},
	{ "BVS",	0102400,	grp4,	0	},
	{ "CCC",	0000257,	grp0,	0	},
	{ "CLC",	0000241,	grp0,	0	},
	{ "CLN",	0000250,	grp0,	0	},
	{ "CLR",	0005000,	grp1,	0	},
	{ "CLRB",	0105000,	grp1,	0	},
	{ "CLV",	0000242,	grp0,	0	},
	{ "CLZ",	0000244,	grp0,	0	},
	{ "CMP",	0020000,	grp2,	0	},
	{ "CMPB",	0120000,	grp2,	0	},
	{ "COM",	0005100,	grp1,	0	},
	{ "COMB",	0105100,	grp1,	0	},
	{ "CSM",	0007000,	grp1,	0	},
	{ "DEC",	0005300,	grp1,	0	},
	{ "DECB",	0105300,	grp1,	0	},
	{ "DIV",	0071000,	grp3,	0	},
	{ "EMT",	0104000,	grp5,	0	},
	{ "HALT",	0000000,	grp0,	0	},
	{ "INC",	0005200,	grp1,	0	},
	{ "INCB",	0105200,	grp1,	0	},
	{ "IOT",	0000004,	grp0,	0	},
	{ "JMP",	0000100,	grp1,	0	},
	{ "JSR",	0004000,	grp3b,	0	},
	{ "MARK",	0006400,	mark,	0	},
	{ "MFPD",	0106500,	grp1,	0	},
	{ "MFPI",	0006500,	grp1,	0	},
	{ "MFPS",	0106700,	grp1,	0	},
	{ "MFPT",	0000007,	grp0,	0	},
	{ "MOV",	0010000,	grp2,	0	},
	{ "MOVB",	0110000,	grp2,	0	},
	{ "MTPD",	0106600,	grp1,	0	},
	{ "MTPI",	0006600,	grp1,	0	},
	{ "MTPS",	0106400,	grp1,	0	},
	{ "MUL",	0070000,	grp3,	0	},
	{ "NEG",	0005400,	grp1,	0	},
	{ "NEGB",	0105400,	grp1,	0	},
	{ "NOP",	0000240,	grp0,	0	},
	{ "RESET",	0000005,	grp0,	0	},
	{ "ROL",	0006100,	grp1,	0	},
	{ "ROLB",	0106100,	grp1,	0	},
	{ "ROR",	0006000,	grp1,	0	},
	{ "RORB",	0106000,	grp1,	0	},
	{ "RTI",	0000002,	grp0,	0	},
	{ "RTS",	0000200,	rts,	0	},
	{ "RTT",	0000006,	grp0,	0	},
	{ "SBC",	0005600,	grp1,	0	},
	{ "SBCB",	0105600,	grp1,	0	},
	{ "SCC",	0000277,	grp0,	0	},
	{ "SEC",	0000261,	grp0,	0	},
	{ "SEN",	0000270,	grp0,	0	},
	{ "SEV",	0000262,	grp0,	0	},
	{ "SEZ",	0000264,	grp0,	0	},
	{ "SOB",	0077000,	sob,	0	},
	{ "SPL",	0000230,	spl,	0	},
	{ "SUB",	0160000,	grp2,	0	},
	{ "SWAB",	0000300,	grp1,	0	},
	{ "SXT",	0006700,	grp1,	0	},
	{ "TRAP",	0104400,	grp5,	0	},
	{ "TST",	0005700,	grp1,	0	},
	{ "TSTB",	0105700,	grp1,	0	},
	{ "TSTSET",	0007200,	grp1,	0	},
	{ "WAIT",	0000001,	grp0,	0	},
	{ "WRTLCK",	0007300,	grp1,	0	},
	{ "XOR",	0074000,	grp3b,	0	},
	{ NULL,		0,		NULL,	0	},
};


error(m, n, p)
char	*m, *n;
{
	char	*fn = Infile? Infile: "(stdin)";

	if (Phase != p) return;
	if (n)
		fprintf(stderr, "BNT11: %s: %d: %s: %s\n", fn, Line, m, n);
	else
		fprintf(stderr, "BNT11: %s: %d: %s\n", fn, Line, m);
	if (++Errors > MAXERRORS) {
		Errors = 0;
		fatal("Too many errors", 0);
	}
}


fatal(m, n)
char	*m, *n;
{
	error(m, n, Phase);
	fprintf(stderr, "Terminating.\n");
	exit(1);
}


findkw(s)
char	*s;
{
	int	i, j, k;

	i = 127;
	j = 64;
	k = 1;
	for(;;) {
		if (	i >= Nkw ||
			s[0] < Kwlist[i].sym[0] ||
			strcmp(s, Kwlist[i].sym) < 0
		) {
			i -= j;
		}
		else if (i < 0 ||
			s[0] > Kwlist[i].sym[0] ||
			strcmp(s, Kwlist[i].sym) > 0
		) {
			i += j;
		}
		else {
			Kw = i;
			return 1;
		}
		j /= 2;
		if (!j) {
			if (!k) return 0;
			k = 0;
			j = 1;
		}
	}
}


printlist() {
	char	*p;

	if (Phase == 2) {
		if (Donumber)
			fprintf(Listfile, "%5d\t", Line-1);
		else
			fprintf(Listfile, "%6o\t", Lastcp);
		Emitbuf[Off] = 0;
		Listbuf[Lp] = 0;
		for (p=Emitbuf; *p; ) {
			if (strlen(p) > 20) {
				fwrite(p, 1, 20, Listfile);
				p += 20;
				while (*p && *p != ' ')
					fputc(*p++, Listfile);
				fputc('\n', Listfile);
				if (Donumber)
					fprintf(Listfile, "%5d\t", Line-1);
				else
					fprintf(Listfile, "\t");
				while (*p == ' ') p++;
			}
			else {
				fprintf(Listfile, "%s", p);
				Off = strlen(p);
				break;
			}
		}
		while (Off++ < 20) fputc(' ', Listfile);
		fputc('\t', Listfile);
		fputs(Listbuf, Listfile);
		Lp = 0;
		Off = 0;
	}
	Lastcp = Cp;
}


rdch() {
	int	c = getc(Input);

	if (feof(Input)) return -1;
	if (Phase == 2 && Dolist) {
		if (Lp >= LISTBUFLEN-1) fatal("List buffer overflow", 0);
		Listbuf[Lp++] = c;
	}
	if (c == '\n') Line++;
	if (Phase == 1) putc(c, Output);
	return c;
}


scan() {
	int	c, g, o, i, r;

	if (Rejected) {
		c = Rejected;
		Rejected = 0;
	}
	else {
		c = rdch();
	}
	while (c == ' ' || c == '\t' || c == '\r' || c == '\f')
		c = rdch();
	if (c == ';') {
		while (!feof(Input) && c != '\n')
			c = rdch();
	}
	if (feof(Input)) {
		strcpy(Text, "<EOF>");
		return ENDFILE;
	}
	Text[0] = c;
	Text[1] = 0;
	if (c == '@') return AT;
	if (c == '(') return LPAREN;
	if (c == ')') return RPAREN;
	if (c == ',') return COMMA;
	if (c == '+') return PLUS;
	if (c == '-') return MINUS;
	if (c == ':') return COLON;
	if (c == '#') return HASH;
	if (c == '=') return EQUAL;
	if (c == '\n') {
		strcpy(Text, "<NL>");
		return NEWLINE;
	}
	if (isdigit(c) || c == '^') {
		Value = 0;
		g = 0;
		o = 0;
		i = 0;
		r = Radix;
		if (c == '^') {
			c = rdch();
			Text[i++] = '^';
			Text[i++] = c;
			switch (c) {
			case 'b': case 'B':	r = 2; break;
			case 'o': case 'O':	r = 8; break;
			case 'd': case 'D':	r = 10; break;
			case 'c': case 'C':	o = 1; break;
			case 'n': case 'N':	g = 1; break;
			default:
				Text[i] = 0;
				error("Bad modifier", Text, 1);
			}
			c = rdch();
		}
		while (isdigit(c)) {
			Text[i++] = c;
			Value = Value * r + c-'0';
			c = rdch();
		}
		Text[i] = 0;
		if (g) Value = -Value;
		if (o) Value = ~Value;
		Rejected = c;
		return NUMBER;
	}
	if (c == '\'') {
		Value = rdch();
		Text[1] = Value;
		Text[2] = '\'';
		Text[3] = 0;
		if (rdch() != '\'') error("Missing '''", NULL, 1);
		return NUMBER;
	}
	if (isalpha(c) || c == '_' || c == '.') {
		i = 0;
		while (isalpha(c) || isdigit(c) || c == '_' || c == '.') {
			if (islower(c)) c = toupper(c);
			if (i >= SYMLEN) {
				Text[i] = 0;
				fatal("Symbol too long", Text);
			}
			Text[i++] = c;
			c = rdch();
		}
		Rejected = c;
		Text[i] = 0;
		if (Text[0] == 'R' && isdigit(Text[1]) && !Text[2]) {
			Reg = Text[1]-'0';
			if (Reg > 7) error("Invalid register", Text, 1);
			return REGISTER;
		}
		if (Text[0] == 'P' && Text[1] == 'C' && !Text[2]) {
			Reg = PC;
			return REGISTER;
		}
		if (findkw(Text)) return KEYWORD;
		return SYMBOL;
	}
	if (c == '"') {
		i = 0;
		c = rdch();
		while (c != '"') {
			if (i >= SYMLEN) {
				Text[i] = 0;
				fatal("Symbol too long", Text);
			}
			if (feof(Input)) fatal("EOF in string", NULL);
			Text[i++] = c;
			c = rdch();
		}
		Text[i] = 0;
		return STRING;
	}
	if (c >= ' ' && c <= '~')
		sprintf(Text, "0%03o (%c)", c, c);
	else
		sprintf(Text, "0%03o", c);
	fatal("Bad input character", Text);
}


synch() {
	int	c;

	c = rdch();
	while (c != '\n') {
		if (feof(Input)) fatal("EOF in error recovery", NULL);
		c = rdch();
	}
}


match(t, m)
char	*m;
{
	if (Token == t) {
		Token = scan();
		return 0;
	}
	else {
		error(m, Text, 1);
		synch();
		return -1;
	}
}


emit(byte) {
	if (Off + 6 > EMITBUFLEN) fatal("Emit buffer overflow", 0);
	if (Phase == 2) {
		if (Doemit) putc(byte, Output);
		if (Dolist && Phase == 2) {
			if (Off > 0) Emitbuf[Off++] = ' ';
			sprintf(&Emitbuf[Off], "%03o", byte & 0377);
			Off += 3;
		}
	}
	if (Doemit) Cp++;
}


emitw(word) {
	if (Off + 9 > EMITBUFLEN) fatal("Emit buffer overflow", 0);
	if (Phase == 2) {
		if (Doemit) {
			putc(word & 0377, Output);
			putc((word >> 8) & 0377, Output);
		}
		if (Dolist && Phase == 2) {
			if (Off > 0) Emitbuf[Off++] = ' ';
			sprintf(&Emitbuf[Off], "%06o", word & 0177777);
			Off += 6;
		}
	}
	if (Doemit) Cp += 2;
}


string(z) {
	char	*p = Text;

	while (*p) emit(*p++);
	match(STRING, "String expected");
	if (z) emit(0);
}


block(word) {
	int	n = Value;

	if (match(NUMBER, "Size value expected")) return;
	if (word) {
		while (n--) { emitw(0); }
	}
	else {
		while (n--) emit(0);
	}
}


data(word) {
	int	a;

	for (;;) {
		if (Token == NUMBER) {
			if (!word && (Value < -0200 || Value > 0377))
				error("Value out of range", Text, 1);
			if (word)
				emitw(Value);
			else
				emit(Value);
			Token = scan();
		}
		else if (Token == SYMBOL) {
			a = findsym(Text, 1);
			if (!word && (a < -0200 || a > 0377))
				error("Value out of range", Text, 1);
			if (word)
				emitw(a);
			else
				emit(a);
			Token = scan();
		}
		else {
			error("Initializer expected", Text, 1);
			synch();
			return;
		}
		if (Token != COMMA) break;
		Token = scan();
	}
}


align(odd) {
	if (odd) {
		if (!(Cp & 1)) emit(0);
	}
	else {
		if (Cp & 1) emit(0);
	}
}


setsym() {
	int	a;

	if (Phase != 1) {
		Token = scan();
		return;
	}
	if (Token == NUMBER) {
		a = Value;
	}
	else if (Token == SYMBOL) {
		a = findsym(Text, 1);
		if (Nosym) error("Undefined symbol", Text, 1);
	}
	else {
		error("Address expected", Text, 1);
		synch();
		return;
	}
	if (Symtab == NULL) error("No symbols present", NULL, 1);
	if (Symtab->flags & F_PREDEF) a = Symtab->addr;
	Symtab->addr = a;
	Token = scan();
}


setup(pval)
int	*pval;
{
	int	a;

	if (Token == NUMBER) {
		a = Value;
	}
	else if (Token == SYMBOL) {
		a = findsym(Text, 1);
		if (Nosym) error("Undefined symbol", Text, 1);
	}
	else {
		error("Address expected", Text, 1);
		synch();
		return;
	}
	Token = scan();
	*pval = a;
}


begincond() {
	int	a1, a2, flag;

	if (Csp >= MAXCOND) fatal("Too many nested .IF's", NULL);
	for (flag=1;;) {
		a1 = address();
		if (match(EQUAL, "Missing '='")) return;
		a2 = address();
		flag &= (a1 == a2);
		if (Token != COMMA) break;
		Token = scan();
	}
	Condstack[Csp++] = Doemit;
	Doemit = flag;
}


endcond() {
	if (!Csp) fatal(".ENDIF without matching .IF", NULL);
	Doemit = Condstack[--Csp];
}


setorigin() {
	Lastcp = Cp;
	if (Bootmode) {
		if (Origin != 0)
			error("Multiple origins in boot code", NULL, 1);
		Origin = Cp;
	}
}


pseudo() {
	int	k = Kw;
	Token = scan();
	switch(Kwlist[k].opcode) {
	case P_ASCII:	string(0); break;
	case P_ASCIZ:	string(1); break;
	case P_BLKB:	block(0); break;
	case P_BLKW:	block(1); break;
	case P_BYTE:	data(0); break;
	case P_ELSE:	if (!Csp) fatal(".ELSE without matching .IF", NULL);
			Doemit = !Doemit; break;
	case P_ENDIF:	endcond(); break;
	case P_EQU:	setsym(); break;
	case P_EVEN:	align(0); break;
	case P_IF:	begincond(); break;
	case P_ODD:	align(1); break;
	case P_ORG:	setup(&Cp); setorigin(); break;
	case P_RADIX:	if (!strcmp(Text, "10")) Value = 10;
			setup(&Radix); break;
	case P_WORD:	data(1); break;
	}
}


struct syment *_findsym(s)
char	*s;
{
	struct syment	*y;

	Nosym = 0;
	for (y = Symtab; y != NULL; y = y->next) {
		if (s[0] == y->name[0] && !strcmp(s, y->name))
			return y;
	}
	Nosym = 1;
	return NULL;
}


findsym(s, rpt)
char	*s;
{
	struct syment	*y;

	if (!Doemit) return -1;
	y = _findsym(s);
	if (Nosym) {
		if (rpt) error("Undefined symbol", s, 2);
		return -1;
	}
	else {
		return y->addr;
	}
}


addsym(sym, addr)
char	*sym;
{
	struct syment	*y;
	char		*s;
	int		flags = 0;

	if (Phase == 2 || !Doemit) return;
	y = _findsym(sym);
	if (!Nosym) {
		if (y->flags & F_PREDEF) {
			addr = y->addr;
			flags = F_PREDEF;
		}
		else {
			error("Duplicate symbol", sym, 1);
		}
	}
	if (	(y = (struct syment *) malloc(sizeof(struct syment))) == NULL
		|| (s = (char *) malloc(strlen(sym)+1)) == NULL
	)
		fatal("Out of memory", NULL);
	strcpy(s, sym);
	y->name = s;
	y->addr = addr;
	y->flags = flags;
	y->next = Symtab;
	Symtab = y;
}


address() {
	int	a;

	if (Token == NUMBER)
		a = Value;
	else if (Token == SYMBOL)
		a = findsym(Text, 1);
	Token = scan();
	return a;
}


deferred() {
	int	r;

	Token = scan();
	if (Token != REGISTER) {
		error("Register expected", Text, 1);
		synch();
		return -1;
	}
	r = Reg;
	Token = scan();
	match(RPAREN, "Missing ')'");
	return r;
}


operand(mp, rp, ap)
int	*mp, *rp, *ap;
{
	if (Token == REGISTER) {
		Token = scan();
		*mp = MREGISTER;
		*rp = Reg;
		return 0;
	}
	if (Token == LPAREN) {
		if ((*rp = deferred()) < 0) return -1;
		if (Token == PLUS) {
			Token = scan();
			*mp = MINCREMENT;
			return 0;
		}
		*mp = MDEFERRED;
		return 0;
	}
	if (Token == AT) {
		Token = scan();
		if (Token == REGISTER) {
			Token = scan();
			*mp = MDEFERRED;
			*rp = Reg;
			return 0;
		}
		if (Token == LPAREN) {
			if ((*rp = deferred()) < 0) return -1;
			if (Token == PLUS) {
				Token = scan();
				*mp = MINCREMENTDEF;
				return 0;
			}
			error("Missing '+'", Text, 1);
			synch();
			return -1;
		}
		if (Token == MINUS) {
			Token = scan();
			if (Token != LPAREN) {
				error("Missing '('", Text, 1);
				synch();
				return -1;
			}
			if ((*rp = deferred()) < 0) return -1;
			*mp = MDECREMENTDEF;
			return 0;
		}
		if (Token == SYMBOL || Token == NUMBER) {
			*ap = address();
			if (Token == LPAREN) {
				if ((*rp = deferred()) < 0) return -1;
				*mp = MINDEXDEF;
				if (*rp == PC) *ap -= Cp + 4;
			}
			else {
				*mp = MINDEXDEF;
				*ap -= Cp + 4;
				*rp = PC;
			}
			return 0;
		}
		if (Token == HASH) {
			Token = scan();
			if (Token != SYMBOL && Token != NUMBER) {
				error("Address expected", Text, 1);
				synch();
				return -1;
			}
			*ap = address();
			*mp = MINCREMENTDEF;
			*rp = PC;
			return 0;
		}
		error("Syntax error", Text, 1);
		synch();
		return -1;
	}
	if (Token == MINUS) {
		Token = scan();
		if ((*rp = deferred()) < 0) return -1;
		*mp = MDECREMENT;
		return 0;
	}
	if (Token == HASH) {
		Token = scan();
		if (Token != SYMBOL && Token != NUMBER) {
			error("Address expected", Text, 1);
			synch();
			return -1;
		}
		*ap = address();
		*mp = MINCREMENT;
		*rp = PC;
		return 0;
	}
	if (Token == SYMBOL || Token == NUMBER) {
		*ap = address();
		if (Token == LPAREN) {
			if ((*rp = deferred()) < 0) return -1;
			*mp = MINDEX;
			if (*rp == PC) *ap -= Cp + 4;
			return 0;
		}
		if (Absmode) {
			/*
		 	* Non-DEC convention
		 	*/
			*mp = MABSOLUTE;
			*rp = PC;
		}
		else {
			*mp = MRELATIVE;
			*rp = PC;
			*ap -= Cp + 4;
		}
		return 0;
	}
	error("Syntax error", Text, 1);
	synch();
	return -1;
}


grp0() {
	int	k = Kw;

	Token = scan();
	emitw(Kwlist[k].opcode);
}


emitoper(m, r, a) {
	if (	m == MINDEX || m == MINDEXDEF ||
		(m == MIMMEDIATE && r == PC) ||
		(m == MABSOLUTE && r == PC)
	)
		emitw(a);
}


grp1() {
	int	i, k = Kw;

	Token = scan();
	if (operand(&Mode1, &Reg1, &Addr1) < 0) return;
	i = Kwlist[k].opcode | (Mode1<<3) | Reg1;
	emitw(i);
	emitoper(Mode1, Reg1, Addr1);
	if (Token == COMMA) {
		error("Too many operands", Text, 1);
		synch();
	}
}


fix2rel() {
	if (	(Mode1 == MINDEX || Mode1 == MINDEXDEF ||
		((Mode1 == MINCREMENT || Mode1 == MINCREMENTDEF) &&
		  Reg1 == PC))
		&&
		(Mode2 == MINDEX || Mode2 == MINDEXDEF ||
		((Mode2 == MINCREMENT || Mode2 == MINCREMENTDEF) &&
		  Reg2 == PC))
	) {
		if (Mode2 == MRELATIVE && Reg2 == PC) Addr2 -= 2;
	}
}


grp2() {
	int	i, k = Kw;

	Token = scan();
	if (operand(&Mode1, &Reg1, &Addr1) < 0) return;
	if (Token != COMMA) {
		error("Too few operands", Text, 1);
		synch();
		return;
	}
	Token = scan();
	if (operand(&Mode2, &Reg2, &Addr2) < 0) return;
	fix2rel();
	i = Kwlist[k].opcode | (Mode1<<9) | (Reg1<<6) | (Mode2<<3) | Reg2;
	emitw(i);
	emitoper(Mode1, Reg1, Addr1);
	emitoper(Mode2, Reg2, Addr2);
}


grp3() {
	int	i, k = Kw;

	Token = scan();
	if (operand(&Mode1, &Reg1, &Addr1) < 0) return;
	if (Token != COMMA) {
		error("Too few operands", Text, 1);
		synch();
		return;
	}
	Token = scan();
	if (operand(&Mode2, &Reg2, &Addr2) < 0) return;
	if (Mode2 != MREGISTER)
		error("Destination must be a register", NULL, 1);
	i = Kwlist[k].opcode | (Mode1<<3) | Reg1 | (Reg2<<6);
	emitw(i);
	emitoper(Mode1, Reg1, Addr1);
}


grp3b() {
	int	i, k = Kw;

	Token = scan();
	if (operand(&Mode1, &Reg1, &Addr1) < 0) return;
	if (Token != COMMA) {
		error("Too few operands", Text, 1);
		synch();
		return;
	}
	Token = scan();
	if (operand(&Mode2, &Reg2, &Addr2) < 0) return;
	if (Mode1 != MREGISTER)
		error("Source/link must be a register", NULL, 1);
	i = Kwlist[k].opcode | (Mode2<<3) | Reg2 | (Reg1<<6);
	emitw(i);
	emitoper(Mode2, Reg2, Addr2);
}


grp4() {
	int	off, k = Kw;

	Token = scan();
	if (Token == NUMBER) {
		off = Value;
		Token = scan();
	}
	else if (Token == SYMBOL) {
		off = (findsym(Text, 1)-Cp-2) / 2;
		Token = scan();
	}
	else {
		error("Offset or destination expected", Text, 1);
		synch();
		off = 0;
	}
	if (Doemit && (off < -0200 || off > 0177))
		error("Branch out of range", NULL, 2);
	emitw((off & 0377) | Kwlist[k].opcode);
}


grp5() {
	int	tc, k = Kw;

	Token = scan();
	tc = Value;
	if (Token != NUMBER) {
		error("Trap code expected", Text, 1);
		synch();
		return;
	}
	if (tc & ~0377) error("Trap code out of range", Text, 1);
	Token = scan();
	emitw((tc & 0377) | Kwlist[k].opcode);
}


mark() {
	int	narg, i, k = Kw;

	Token = scan();
	narg = Value;
	if (Token != NUMBER) {
		error("Argument count expected", Text, 1);
		synch();
		return;
	}
	if (narg & ~077) error("Argument count out of range", Text, 1);
	Token = scan();
	i = Kwlist[k].opcode | narg;
	emitw(i);
}


rts() {
	int	i, k = Kw;

	Token = scan();
	if (Token != REGISTER) {
		error("Register expected", Text, 1);
		synch();
		return;
	}
	i = Kwlist[k].opcode | Reg;
	Token = scan();
	emitw(i);
}


sob() {
	int	i, off, r, k = Kw;

	Token = scan();
	if (Token != REGISTER) {
		error("Register expected", Text, 1);
		synch();
		return;
	}
	r = Reg;
	Token = scan();
	if (Token != COMMA) {
		error("Too few operands", Text, 1);
		synch();
		return;
	}
	Token = scan();
	if (Token == NUMBER) {
		off = Value;
		Token = scan();
	}
	else if (Token == SYMBOL) {
		off = -((findsym(Text, 1)-Cp-2) / 2);
		Token = scan();
	}
	else {
		error("Offset or destination expected", Text, 1);
		synch();
		off = 0;
	}
	if (Doemit && (off < 0 || off > 077))
		error("Branch out of range", NULL, 2);
	off = off & 077;
	i = Kwlist[k].opcode | (r << 6) | off;
	emitw(i);
}


spl() {
	int	pl, i, k = Kw;

	Token = scan();
	pl = Value;
	if (Token != NUMBER) {
		error("Priority level expected", Text, 1);
		synch();
		return;
	}
	if (pl & ~07) error("Priority level out of range", Text, 1);
	Token = scan();
	i = Kwlist[k].opcode & pl;
	emitw(i);
}


skipnewline(req) {
	if (Token == NEWLINE) {
		while (Token == NEWLINE) {
			if (Dolist) printlist();
			Token = scan();
		}
	}
	else if (req) {
		fatal("Newline expected", Text);
		synch();
	}
}


pass() {
	Cp = Lastcp = 0;
	Line = 1;
	Rejected = 0;
	Radix = 8;
	Token = scan();
	while (!feof(Input)) {
		if (Token == NEWLINE)
			;
		else {
			while (Token == SYMBOL) {
				addsym(Text, Cp);
				Token = scan();
				if (match(COLON, "Missing ':'")) break;
				skipnewline(0);
			}
			if (Token != KEYWORD)  {
				error("Instruction expected", Text, 1);
				synch();
			}
			(*Kwlist[Kw].group)();
		}
		skipnewline(1);
	}
}


init() {
	for (Nkw=0; Kwlist[Nkw].sym; Nkw++)
		;
	Input = stdin;
}


assemble() {
	Input = Infile == NULL? stdin: fopen(Infile, "rb");
	if (Input == NULL) fatal("Could not open file", Infile);
	if ((Output = fopen(Tmpfile, "wb")) == NULL) {
		fclose(Input);
		fatal("Could not create tmp file", Tmpfile);
	}
	pass();
	if (Infile) fclose(Input);
	fclose(Output);
	if (Csp) fatal("Missing .ENDIF at EOF", NULL);
	if ((Input = fopen(Tmpfile, "rb")) == NULL)
		fatal("Could not reopen tmp file", Tmpfile);
	Output = Outfile == NULL? stdout: fopen(Outfile, "wb");
	if (Output == NULL) {
		if (!Keeptmp) unlink(Tmpfile);
		fclose(Input);
		fatal("Could create open file", Outfile);
	}
	Phase = 2;
	pass();
	fclose(Input);
	if (Outfile) fclose(Output);
	if (!Keeptmp) unlink(Tmpfile);
}


test() {
	int	i;

	for (i=0; i<Nkw; i++) {
		if (!findkw(Kwlist[i].sym)) {
			fprintf(stderr, "findkw() failed @ %d\n", i);
			exit(1);
		}
	}
	fprintf(stderr, "Passed\n");
	exit(0);
}


usage() {
	printf("\n"
	"Usage: bnt11 [-aklnv] [-o file] [-t file] [-D sym=val] [file]\n"
	"\n"
	"-a      Use absolute mode by default [PC-relative]\n"
	"-b      Generate boot code (ODT input)\n"
	"-k      Keep temporary file\n"
	"-l      Generate list file\n"
	"-n      Generate line numbers in list\n"
	"-v      Print version\n"
	"-o file Direct output to 'file' [%s]\n"
	"-t file Specify temporary file name [%s]\n"
	"-D s=v  Override definition of 's' with value 'd'\n"
	"file    Input file [%s]\n"
	"\n",
	Outfile==NULL? "(stdout)": Outfile,
	Tmpfile,
	Infile==NULL? "(stdin)": Infile);
	exit(1);
}


version() {
	printf("BTN11 -- Version 0.9 -- Nils M Holm, 2001 -- Public Domain\n");
}


openlist() {
	char	listname[256], *p;
	int	k;

	if (!Infile) fatal("The -l option requires a file name", NULL);
	if ((p = strrchr(Infile, '/')) == NULL) p = Infile;
	k = strlen(p);
	if (k + 4 >= 256) fatal("File name too long", p);
	strcpy(listname, p);
	if ((p = strrchr(listname, '.')) == NULL) p = &listname[k];
	strcpy(p, ".lst");
	if ((Listfile = fopen(listname, "w")) == NULL)
		fatal("Could not create list file", listname);
}


genboot() {
	FILE	*in, *out;
	int	x;

	if ((in = fopen("btn11.boot", "rb")) == NULL)
		fatal("Could not reopen work file: btn11.boot", NULL);
	if (Outfile) {
		if ((out = fopen(Outfile, "wb")) == NULL)
			fatal("Could create file", Outfile);
	}
	else {
		out = stdout;
	}
	fprintf(out, "%o/", Origin);
	for (;;) {
		x = fgetc(in);
		if (feof(in)) break;
		x = (fgetc(in) << 8) | x;
		fprintf(out, "%o\n", x);
	}
	fputc('\r', out);
	fclose(in);
	fclose(out);
	if (!Keeptmp) unlink("btn11.boot");
}


defsym(s)
char	*s;
{
	char	*v;

	if ((v = strchr(s, '=')) == NULL) usage();
	*v++ = 0;
	addsym(s, atoi(v));
	Symtab->flags |= F_PREDEF;
}


main(argc, argv)
char	**argv;
{
	int	i, j;
	char	*ofile;

	init();
	for (i=1; i<argc && argv[i][0] == '-'; i++) {
		if (!strcmp(argv[i], "-test")) test();
		for (j=1; argv[i][j]; j++) {
			switch (argv[i][j]) {
			case 'a':	Absmode = 1;
					break;
			case 'b':	Bootmode = 1;
					break;
			case 'k':	Keeptmp = 1;
					break;
			case 'l':	Dolist = 1;
					break;
			case 'n':	Donumber = 1;
					break;
			case 'o':	if (i+1 >= argc) usage();
					Outfile = argv[++i];
					j = strlen(argv[i])-1;
					break;
			case 't':	if (i+1 >= argc) usage();
					Tmpfile = argv[++i];
					j = strlen(argv[i])-1;
					break;
			case 'v':	version();
					break;
			case 'D':	if (i+1 >= argc) usage();
					defsym(argv[++i]);
					j = strlen(argv[i])-1;
					break;
			default: 	usage();
			}
		}
	}
	if (Bootmode) {
		ofile = Outfile;
		Outfile = "btn11.boot";
	}
	if (i < argc) Infile = argv[i++];
	if (i < argc) usage();
	if (Dolist) openlist();
	assemble();
	if (Dolist) fclose(Listfile);
	if (Bootmode) {
		Outfile = ofile;
		genboot();
	}
	exit(Errors? 1: 0);
}
